------------------------------------------------------------------------------
--                                                                          --
--                         GNAT COMPILER COMPONENTS                         --
--                                                                          --
--                               I T Y P E S                                --
--                                                                          --
--                                 S p e c                                  --
--                                                                          --
--                            $Revision: 1.13 $                              --
--                                                                          --
--           Copyright (c) 1992,1993,1994 NYU, All Rights Reserved          --
--                                                                          --
-- GNAT is free software;  you can  redistribute it  and/or modify it under --
-- terms of the  GNU General Public License as published  by the Free Soft- --
-- ware  Foundation;  either version 2,  or (at your option) any later ver- --
-- sion.  GNAT is distributed in the hope that it will be useful, but WITH- --
-- OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY --
-- or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License --
-- for  more details.  You should have  received  a copy of the GNU General --
-- Public License  distributed with GNAT;  see file COPYING.  If not, write --
-- to the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. --
--                                                                          --
------------------------------------------------------------------------------

--  This package contains declarations for handling of implicit types

with Einfo;    use Einfo;
with Sem_Util;
with Types;    use Types;

package Itypes is

   --------------------
   -- Implicit Types --
   --------------------

   --  Implicit types are types created by the semantic or the expander either
   --  by convenience or to stick with the Ada formal semantics. They are
   --  entities which don't appear as a Defining_Identifier of some declaration
   --  Node in order to save tree expension or when there is no Ada syntax for
   --  defining them (the class wide type associated to a tagged type for
   --  instance). They are attached to the tree through the node related
   --  to their creation.

   --  Here is the list of nodes that may contain implicit types:

   --    * N_Full_Type_Declaration:
   --        - Array types  : implicit base type, implicit index types
   --        - Tagged types : class wide type

   --    * N_Private_Type_Declaration:
   --        - Discriminated types : implicit access types

   --    * N_Formal_Type_Declaration
   --    * N_Incomplete_Type_Declaration
   --        - Discriminated types : implicit access types

   --    * N_Subtype_Declaration
   --        - Array types  : implicit index types

   --    * N_Formal_Object_Declaration:
   --    * N_Component_Declaration:
   --    * N_Object_Declaration:
   --        - discriminated or indexed objects: implicit constrained subtype
   --        - anonymous array objects: implicit type and indexes
   --        - dispatch table object: anonymous access to primitive types

   --    * N_Loop_Parameter_Specification:
   --        -  implicit range type

   --    * N_Function_Specification:
   --    * N_Procedure_Specification:
   --    * N_Entry_Body:
   --    * N_Accept_Statement:
   --        -  implicit formal access type

   --    * N_Entry_Declaration:
   --        - implicit formal access type
   --        - implicit index (for families)

   --    * N_Freeze_Entity:
   --        - implicit types attached to a N_Object_Declaration may be moved
   --          to their own N_Freeze node when this one is inserted before the
   --          object declaration

   --    * N_Implicit_Types:

   --        These nodes are introduced when it is not possible to attach an
   --        implicit type to the related syntax node because it causes
   --        order of elaboration problems. This situation happens for
   --        object declarations where two potentially contradictory
   --        situations must be handled
   --
   --          X : Uncontrained_Type := Init_Value_Aggregate;
   --          X : Uncontrained_Type (Constraints) := Init_Value_Aggregate;
   --
   --        in the first case the anonymous constained_type depends on the
   --        anonymous aggregate types. In the second case, the dependancy
   --        is the other way around.

   --  Subexpression nodes

   --    * N_Slice:
   --        - implicit index type
   --        - implicit slice Subtype

   --    * N_Aggregate:
   --        - implicit constrained subtypes

   --    * N_String_Literal:
   --        - implicit string subtype

   --    * N_Allocator:
   --        - implicit access type
   --        - explicit constrained designated subtype (this one is not an
   --          implicit type, i.e. the tree is generated for its declaration)
   --          this is a peculiarity.

   --    * N_Reference
   --        - implicit access type

   --    * N_Expression_Actions:
   --        - implicit types attached with the Expression Node are moved up
   --          to the expression action in order to ease Gigi processing.

   --  Note: the set of nodes listed above is covered by the subtype
   --  N_Has_Itypes. Such nodes have a First_Itype field that is either
   --  Empty if there are no associated Itypes, or points to the first
   --  entity on the list of Itypes if there are Itypes. In addition such
   --  nodes have a Has_Dynamic_Itype flag that is set if any of the
   --  entities in the list is for a dynamic implicit type (one which has
   --  variable bounds or other data that must be elaborated).

   --  Note: N_Has_Itypes includes N_Subtype_Indication, even though it is
   --  not listed above. This is for purely technical reasons, to accomodate
   --  the restriction that the node types for which Itypes are present must
   --  have contiguous codes so that N_Has_Itypes can be declared. The two
   --  fields are always unused, i.e. First_Itype is always Empty for a
   --  subtype indication node, and Has_Dynamic_Itype is always False.

   procedure Transfer_Itypes (From : Node_Id; To : Node_Id);
   --  Move the implicit types from node From to node To.

   procedure Attach_Itype_To (N : Node_Id; E : Entity_Id);
   --  Attach the Itype E at the end of the list of Itypes associated with
   --  node N, which must be one of the node types listed above.

   procedure Prepend_Itype_To (N : Node_Id; E : Entity_Id);
   --  Attach the Itype E to the start of the list of Itypes associated
   --  with node N, which must be one of the node types listed above.

   function New_Itype
     (Ekind        : Entity_Kind;
      In_Node      : Node_Id;
      Related_Id   : Entity_Id   := Empty;
      Suffix       : Character   := ' ';
      Suffix_Index : Nat         := 0;
      Scope_Id     : Entity_Id   := Sem_Util.Current_Scope)
      return         Entity_Id;
   --  This is used in cases where there is a new underlying type or subtype
   --  introduced implicitly. Examples include:
   --
   --   a. the implicit unconstrained array type introduced by a
   --      constrained array type declaration.
   --
   --   b. the implicit subtype introduced by a subtype indication given as
   --      "Integer range 1 .. 2".
   --
   --   Related_Id is present only if the implicit type name may be referenced
   --   as a public symbol, and thus needs a unique external name. The name
   --   is created by a call to:
   --
   --     New_External_Name (Chars (Related_Id), Suffix, Suffix_Index, 'T')
   --
   --   If the implicit type does not need an external name, then the
   --   Related_Id parameter is omitted (and hence empty). In this case
   --   Suffix and Suffix_Index are ignored and the implicit type name is
   --   created by a call to New_Internal_Name ('T').
   --
   --   Note that in all cases, the name starts with "T". This is used
   --   to identify implicit types in the error message handling circuits.
   --
   --  The Scope_Id parameter specifies the scope of the created type, and
   --  is normally the Current_Scope as shown, but can be set otherwise.

   function New_Itype_Not_Attached
     (Ekind        : Entity_Kind;
      Loc          : Source_Ptr;
      Related_Id   : Entity_Id := Empty;
      Suffix       : Character := ' ';
      Suffix_Index : Nat       := 0;
      Scope_Id     : Entity_Id := Sem_Util.Current_Scope)
      return         Entity_Id;
   --  Same function as above but the Itype it not attached directly to a Node
   --  This is useful when the attachment to the node cannot be performed when
   --  the Itype is created. The Itype must hen be attached later on by a call
   --  to Attach_Itype_To.

   -----------------------------------------------
   -- Handling of Actions Associated with Nodes --
   -----------------------------------------------

   --  The evaluation of certain expression nodes involves the elaboration
   --  of associated types and other declarations, and the execution of
   --  statement sequences. The tree node N_Expression_Actions is used in
   --  the tree to represent this semantic requirement. However these nodes
   --  appear only temporarily in the tree, since if code is being generated,
   --  Gigi expects that all expression actions nodes must be removed from
   --  the tree, since it does not have the capability of handling complex
   --  declarations in the middle of evaluation of an expression.

   --  This means that the expansion routine for expression actions nodes must
   --  find an appropriate place in the tree to rehang the actions so that they
   --  will be evaluated at the appropriate point.

   --  Some cases are simple:

   --    For an expression occurring in a simple statement that is in a list
   --    of statements, the actions are simply inserted into the list before
   --    the associated statement.

   --    For an expression occurring in a declaration (declarations always
   --    appear in lists), the actions are similarly inserted into the list
   --    just before the associated declaration.

   --  The following special cases arise:

   --    For actions associated with the right operand of a short circuit
   --    form, the actions are first stored in the short circuit form node
   --    in the Actions field. The expansion of these forms subsequently
   --    expands the short circuit forms into if statements which can then
   --    be moved as described above.

   --    For actions appearing in the Condition expression of a while loop,
   --    or an elsif clause, the actions are similarly temporarily stored in
   --    in the node (N_Elsif_Part or N_Iteration_Scheme) associated with
   --    the expression using the Condition_Actions field. Sunsequently, the
   --    expansion of these nodes rewrites the control structures involved to
   --    reposition the actions in normal statement sequence.

   --    For actions appearing in the then or else expression of a conditional
   --    expression, these actions are similarly placed in the node, using the
   --    Then_Actions or Else_Actions field as appropriate. Once again the
   --    expansion of the N_Conditional_Expression node rewrites the node so
   --    that the actions can be normally positioned.

   --  Basically what we do is to climb up to the tree looking for the proper
   --  insertion point, as described by one of the above cases.

   --  In addition to the actions of expression actions nodes, the definition
   --  of implicit types (i.e. the N_Implicit_Type node) must also be inserted
   --  in the tree in a similar manner.

   procedure Insert_Action  (Action  : Node_Id; Node : Node_Id);
   procedure Insert_Actions (Actions : List_Id; Node : Node_Id);
   --  These routines are used to promote the a single action or actions
   --  to the top level. Typical actions to be promoted in this way are
   --  the declaration of an Itype, or the actions of an expression actions
   --  node. For an exact definition of what top-level means, see body. The
   --  second argument is the node to which the action is originally attached.

end Itypes;
